@@ -4,7 +4,6 @@ State tracking functionality for django models
 """
 import inspect
 import sys
-import warnings
 from functools import wraps
 
 from django.db import models
@@ -65,6 +64,10 @@ class TransitionNotAllowed(Exception):
         super(TransitionNotAllowed, self).__init__(*args, **kwargs)
 
 
+class InvalidResultState(Exception):
+    """Raised when we got invalid result state"""
+
+
 class ConcurrentTransition(Exception):
     """
     Raised when the transition cannot be executed because the
@@ -91,11 +94,7 @@ class Transition(object):
         if not self.permission:
             return True
         elif callable(self.permission):
-            try:
-                return bool(self.permission(instance, user))
-            except TypeError:
-                warnings.warn('Callable permissions without instance parameter is depricated')
-                return bool(self.permission(user))
+            return bool(self.permission(instance, user))
         elif user.has_perm(self.permission, instance):
             return True
         elif user.has_perm(self.permission):
@@ -114,12 +113,8 @@ def get_available_FIELD_transitions(instance, field):
 
     for name, transition in transitions.items():
         meta = transition._django_fsm
-
-        for state in [curr_state, '*', '+']:
-            if state in meta.transitions:
-                transition = meta.transitions[state]
-                if all(map(lambda condition: condition(instance), transition.conditions)):
-                    yield transition
+        if meta.has_transition(curr_state) and meta.conditions_met(instance, curr_state):
+            yield meta.get_transition(curr_state)
 
 
 def get_all_FIELD_transitions(instance, field):
@@ -322,6 +317,10 @@ class FSMFieldMixin(object):
         try:
             result = method(instance, *args, **kwargs)
             if next_state is not None:
+                if hasattr(next_state, 'get_state'):
+                    next_state = next_state.get_state(
+                        instance, transition, result,
+                        args=args, kwargs=kwargs)
                 self.set_proxy(instance, next_state)
                 self.set_state(instance, next_state)
         except Exception as exc:
@@ -548,6 +547,39 @@ def has_transition_perm(bound_method, user):
     im_self = getattr(bound_method, 'im_self', getattr(bound_method, '__self__'))
     current_state = meta.field.get_state(im_self)
 
-    return (meta.has_transition(current_state)
-            and meta.conditions_met(im_self, current_state)
-            and meta.has_transition_perm(im_self, current_state, user))
+    return (meta.has_transition(current_state) and
+            meta.conditions_met(im_self, current_state) and
+            meta.has_transition_perm(im_self, current_state, user))
+
+
+class State(object):
+    def get_state(self, model, transition, result, args=[], kwargs={}):
+        raise NotImplementedError
+
+
+class RETURN_VALUE(State):
+    def __init__(self, *allowed_states):
+        self.allowed_states = allowed_states if allowed_states else None
+
+    def get_state(self, model, transition, result, args=[], kwargs={}):
+        if self.allowed_states is not None:
+            if result not in self.allowed_states:
+                raise InvalidResultState(
+                    '{} is not in list of allowed states\n{}'.format(
+                        result, self.allowed_states))
+        return result
+
+
+class GET_STATE(State):
+    def __init__(self, func, states=None):
+        self.func = func
+        self.allowed_states = states
+
+    def get_state(self, model, transition, result, args=[], kwargs={}):
+        result_state = self.func(model, *args, **kwargs)
+        if self.allowed_states is not None:
+            if result_state not in self.allowed_states:
+                raise InvalidResultState(
+                    '{} is not in list of allowed states\n{}'.format(
+                        result, self.allowed_states))
+        return result_state
